
# GNU HEADER HERE!

##########################################################################################
#
#	DUPLICATES.PM - metods exported:
#
#	TO DO!
#
##########################################################################################



package EPrints::Duplicates;


use strict;
use EPrints::User;
use EPrints::Database;
use EPrints::Session;
use EPrints::EPrint;
use EPrints::DataSet;


##SEB this module manages the discovery of duplicates etc...



# For later:

#@EPrints::Duplicates::VALID_DATASETS = ( "archive" );	# for now... include buffer at some point

#%EPrints::Duplicates::SPECIAL_FIELDS = (
#	article=>"publication",
#	book=>"publisher",
#	book_section=>"title",
#	conference_item=>"event_title",
#	monograph=>"publisher",
#	patent=>"id_number",
#	thesis_type=>"department",
#	);
	
	
@EPrints::Duplicates::AUTHORS_TYPE = ( "creators", "editors", "exhibitors" );	

$EPrints::Duplicates::THRESHOLD = 60;		# 60% minimum to be considered duplicates

$EPrints::Duplicates::WEIGHT_TRIGRAMS = 0.7;	# 70% of total score goes to the trigrams comparaison
$EPrints::Duplicates::WEIGHT_AUTHORS = 0.3;	# 30% of total score goes to the authors comparaison

# those words are removed when generating trigrams:
@EPrints::Duplicates::EXCLUDED_WORDS = ( 'the', 'or', 'in', 'with', 'for', 'a', 'an', 'on', 'up', 'to', 
		'into', 'upto', 'upon', 'does', 'do', 'have', 'has', 'is', 'are', 'there', 'and', 'of', 'at', 'from',
		'it', 'they');

$EPrints::Duplicates::TRIGRAMS_SOURCE = "title";	# which field is used to generate trigrams from


###	
###	OK: but this might use loads of resources :s
###
#return a ref to a hash: $results{eprintid} = score
# NOT USED!
sub find_duplicates
{
	
	my ( $session, $eprintid ) = @_;
	
	return undef if ( !defined $session || !defined $eprintid );

	my $fingerprints = get_all_fingerprints( $session );	# format: $$fingerprints{eprintid} = fingerprint
	
	my $reference;
	
	if( !defined $$fingerprints{$eprintid} )
	{
		# it generates it, but does not save it to the DB!!
		$reference = generate_fingerprint_by_id( $session, $eprintid );
		
		return undef if ( !defined $reference );
	}
	else
	{
		$reference = $$fingerprints{$eprintid};
	}
	
	my %results = ();
	
	my ( $id, $score );
	
	foreach $id ( keys %$fingerprints )
	{
		next if ( $id eq $eprintid );
	
		$score = score( $reference, $$fingerprints{$id} );
		
		if( $score >= $EPrints::Duplicates::THRESHOLD )
		{
			$results{$id} = $score;
		}
		
	}
	
	return \%results;

}


# OK
sub empty_report_table
{
	my ( $session, $keep_ignored ) = @_;
	
	if(!defined $keep_ignored || $keep_ignored)
	{
		return $session->get_db->do( "DELETE FROM duplicates_report WHERE ignored=\'FALSE\';");
	}
	else
	{
		return $session->get_db->do( "TRUNCATE TABLE duplicates_report;" );
	}
}



###	
###	OK
###
sub remove_from_ignore_list
{
	my ( $session, $ref, $dup )  = @_;
	
	my $sql = "UPDATE duplicates_report SET ignored=\'FALSE\' WHERE eprintid_ref=$ref AND eprintid_dup=$dup;";

	return $session->get_db->do( $sql );
}

###	
###	OK
###
sub add_to_ignore_list
{
	my ( $session, $ref, $dup )  = @_;
	
	my $sql = "UPDATE duplicates_report SET ignored=\'TRUE\' WHERE eprintid_ref=$ref AND eprintid_dup=$dup;";

	return $session->get_db->do( $sql );
}




###	
###	IMPROVE OR REMOVE
###
# return some very basic stats on how many suspected duplicates have been found etc.
sub get_stats
{
	my ( $session ) = @_;
	
	my %stats;	# will return a ref to this hash
	
	$stats{dup_total_count} = 0;
	$stats{dup_ignored} = 0;
	$stats{dup_non_ignored} = 0;
# maybe more stats....
	
	if(!defined $session)
	{
		return \%stats;
	}

	my $sql = "SELECT COUNT(score) FROM duplicates_report;";
	
	my $sth = $session->get_db->prepare( $sql );
	if( $session->get_db->execute( $sth, $sql ) )
	{
		my @row = $sth->fetchrow;

		if(defined $row[0])
		{
			$stats{dup_total_count} = $row[0];
		}
	}
	
	$sql = "SELECT COUNT(score) FROM duplicates_report WHERE ignored=\'TRUE\';";
	$sth = $session->get_db->prepare( $sql );
	
	if( $session->get_db->execute( $sth, $sql ) )
	{
		my @row = $sth->fetchrow;

		if(defined $row[0])
		{
			$stats{dup_ignored} = $row[0];
		}
	}
	
	$stats{dup_non_ignored} = $stats{dup_total_count} - $stats{dup_ignored};
	

# add how many fingerprints we have too:

	$sql = "SELECT COUNT(eprintid) FROM duplicates_archive_fingerprints;";
	
	$sth = $session->get_db->prepare( $sql );
	if( $session->get_db->execute( $sth, $sql ) )
	{
		my @row = $sth->fetchrow;

		if(defined $row[0])
		{
			$stats{fingerprints} = $row[0];
		}
	}

	return \%stats;
}




# retrieve duplicates from duplicates_report for the webpage....
sub retrieve_duplicates
{
	my ( $session, $offset, $limit ) = @_;
	
	return undef if( !defined $session || !defined $offset || !defined $limit );
	
	my $sql = "SELECT eprintid_ref, eprintid_dup, score, ignored FROM duplicates_report WHERE ignored=\'FALSE\' ORDER BY score DESC LIMIT $offset, $limit;";

	my $sth = $session->get_db->prepare($sql);

	if(!$session->get_db->execute($sth,$sql))
	{
		return undef;		# is that cool?
	}

	my @dups;

	while( my @row = $sth->fetchrow )
	{
		push @dups, { eprintid_ref=> $row[0], eprintid_dup=>$row[1],score=>$row[2] };
	}
	
	return \@dups;

}



# retrieve duplicates from duplicates_report for the webpage....
sub retrieve_ignored_duplicates
{
	my ( $session, $offset, $limit ) = @_;
	
	return undef if( !defined $session || !defined $offset || !defined $limit );
	
	my $sql = "SELECT eprintid_ref, eprintid_dup, score, ignored FROM duplicates_report WHERE ignored = \'TRUE\' ORDER BY score DESC LIMIT $offset, $limit;";

	my $sth = $session->get_db->prepare($sql);

	if(!$session->get_db->execute($sth,$sql))
	{
		return undef;		# is that cool?
	}

	my @dups;

	while( my @row = $sth->fetchrow )
	{
		push @dups, { eprintid_ref=> $row[0], eprintid_dup=>$row[1], score=>$row[2] };
	}
	
	return \@dups;

}






sub generate_fingerprint_by_id
{
	my ( $session, $eprintid ) = @_;
	
	my $eprint = new EPrints::EPrint( $session, $eprintid );
	
	return undef if ( !defined $eprint );
	
	return generate_fingerprint( $eprint );
}



###	
###	OK?
###
sub generate_fingerprint
{
	my ($eprint) = @_;

	my $text = $eprint->get_value( $EPrints::Duplicates::TRIGRAMS_SOURCE );


#	MAY WANT TO ENABLE THAT SECTION FOR MAYBE BETTER RESULTS (UNTESTED):
#
#	my $type = $eprint->get_value("type");
#	
#	my $special_field = $EPrints::Duplicates::SPECIAL_FIELDS{$type};
#	
#	if(defined $special_field)
#	{
#		my $value = $eprint->get_value( "$special_field" );
#		if(defined $value)
#		{
#			$text .= " $value";
#		}
#	}
	
	
	my $trigrams = "";
	my $is_last = 0;
	
	if(defined $text)
	{
		my @words = split(' ', $text);
		my @selected = ();
		
		my ($w, $ex);
		
		foreach $w (@words)
		{
			# the next 3 loops filter out some unneeded characters at the beginning and end of words
			
			next if(!defined $w);
			
			while($w =~ /[,:.;'#~\"\\\/\!\?\(\)\_\{\}\[\]\ ]$/)
			{
				chop($w);
			}
			
			while($w =~ /^[\(\)\-\_\{\}\[\]\=\ ]/)
			{
				$w =substr($w,1,length($w)-1);
			}
			
			if($w =~ /^(.*)\/(.*)$/)
			{
				push @words, $2;
				$w = $1;
				# dirt :-)
			}
			
			# review use of $bool:
			$is_last = 0;
			
			next if (length($w) < 2 && $w !~ /\d+/);	# we ignore words which length is 1 char unless it's a number
			
			$w = uc $w;
			
			foreach $ex ( @EPrints::Duplicates::EXCLUDED_WORDS )
			{
				$is_last++ if( $w eq uc($ex) );
				
				last if $is_last;
			}
			
			next if $is_last;
			
			$is_last = 0;
			
			# checks if already selected (don't need word twice...)
			foreach (@selected)
			{
				$is_last++ if( $_ eq $w );
				last if $is_last;
			}
			
			next if $is_last;
			
			push @selected, $w;
			
		}

		# generate the trigrams from the list of selected words:
		$trigrams = generate_trigrams( join(' ', @selected) );
	}
	else
	{
		$trigrams = "0";
	}

# add type + editors/creators/exhibitors at the end of the trigrams...

	$trigrams.="#".$eprint->get_value("type");
	
	$trigrams.="#".join(":", @{get_authors($eprint)});
	
	return $trigrams;

# at the end, a fingerprint looks like:
# L:AAA:AAB:ABB:ABC#eprint_type#author1:author2:author3
# L: number of trigrams (we calculate this value here, so we don't have to do it in the scoring method... remember the scoring method has to be quick... generation of
# fingerprints can be slow, it doesn't matter)

}


###	
###	OK
###
sub get_authors
{
	my ( $eprint ) = @_;
	
	my (@names, @people, $p, %already_done, $name, $type);
	
	
	foreach $type ( @EPrints::Duplicates::AUTHORS_TYPE )
	{

		@people = @{$eprint->get_value( $type )};
	
		foreach $p ( @people )
		{
			$name = lc $p->{main}->{family};
			
			if( !defined $already_done{$name} )
			{		
				push @names, $name;	
				$already_done{$name} = 1;
			}
		}
			
	}
	
	return \@names;
}



# this sub generates trigrams from a list of words, this sub shouldn't be called outside the module
###	
###	OK
###
sub generate_trigrams
{
	my ( $string ) = @_;
	
	$string = "  ".$string."  ";		# padding: more trigrams = more precision!

	my $len = length( $string );
	
	my ( %trigrams, $i );
	
	for( $i=0; $i<$len-2; $i++ )
	{
		my $tri = substr($string, $i, 3);
		
		$trigrams{$tri} = 1;
	}
	
	return $len.":".join(':', keys %trigrams);
	
}




###	
###	OK... can still be improved to have quicker results...
###
sub score
{
	my ( $fp1, $fp2 ) = @_;		# <= update 
	# fingerprint1, fingerprint2

	if(!defined $fp1 || !defined $fp2)
	{
		return 0;
	}

# the fingerprints should look like: L:AAA:ABB:BBC#eprints_type#auth1:auth2:auth3
# so if everything's ok, it looks like:
#	$segments[0] => trigrams
#	$segments[1] => eprints type
# 	$segments[2] => list of authors' family name

	my @segments1 = split("#", $fp1);
	my @segments2 = split("#", $fp2);
	
	my (@trigrams1,@trigrams2,@authors1,@authors2, $type1, $type2, $len1, $len2);
	
	
	@trigrams1 = split(':', $segments1[0]) if (defined $segments1[0] );
	@trigrams2 = split(':', $segments2[0]) if(defined $segments2[0]);
	
	$type1 = $segments1[1];
	$type2 = $segments2[1];
	
	if($type1 ne $type2)
	{
		return 0;		# new scoring method!
	}


	@authors1 = split(":", $segments1[2]) if (defined $segments1[2]);
	
	$len1 = shift @trigrams1;	# len is actually the number of trigrams

	my %h1;
	foreach(@trigrams1)
	{
		$h1{$_}=1;
	}
	

	@authors2 = split(":", $segments2[2]) if (defined $segments2[2]);
	
	$len2 = shift @trigrams2;	# len is actually the number of trigrams
	
	my %h2;
	foreach(@trigrams2)
	{
		$h2{$_}=1;
	}
	

	my $matches = 0;
	
	foreach(keys %h1)
	{
		if(defined $h2{$_})
		{
			$matches++;
			$h2{$_}++;	# mark as seen
		}
	}
		
	foreach(keys %h2)
	{
		if($h2{$_} < 2 && defined $h1{$_})
		{
			$matches++;
		}
	}
	
	
# the two scores that merge into one global score
	my ($score,$score_tri, $score_authors);

	$score_tri = ($matches/($len1+$len2-4-$matches));
	
	$score_tri = 1 if $score_tri > 1;
	
	$score_authors = score_authors( \@authors1, \@authors2 );
	
	$score = ($score_tri * $EPrints::Duplicates::WEIGHT_TRIGRAMS + $score_authors * $EPrints::Duplicates::WEIGHT_AUTHORS)*100;
	
	return int( $score );	# score is now a percentage, not a value between 0 and 1
}


###	
###	OK...
###
sub score_authors
{
	my ( $authors1, $authors2 ) = @_;
	
	my ($big_array, $small_array);
	
	if(scalar(@$authors1) == 0 || scalar(@$authors2) == 0)
	{
		return 0;
	}
	
	if( scalar(@$authors1) > scalar(@$authors2) )
	{
		$big_array = $authors1;
		$small_array = $authors2;
	}
	else
	{
		$big_array = $authors2;
		$small_array = $authors1;
	}
	
	my ($e1, $e2);
	
	my $sim = 0;
	
	foreach $e1 ( @$big_array )
	{
		foreach $e2 ( @$small_array )
		{
			if( $e1 eq $e2 )
			{
				$sim ++;
			}
		}
	}
	
	$sim /= scalar( @$big_array );	# scalar( @$big_array ) won't be equal to 0 since we check that a few lines above ;)
	
	return $sim;
}



# return a ref to an array/hash of fingerprints...
sub get_all_fingerprints
{
	my ( $session ) = @_;

	my %fp;
	
	my $sql = "SELECT eprintid, fingerprint from duplicates_archive_fingerprints;";
	
	my $sth = $session->get_db->prepare( $sql );
	if( $session->get_db->execute( $sth, $sql ) )
	{
		while( my @row = $sth->fetchrow )
		{
			
			$fp{$row[0]} = $row[1];	# $fp{ eprintid } = fingerprint;
			
		}
	
	}
	#else
	#{
	#	big trouble!
	#}
	
	return \%fp;
		
}



# UNUSED?
sub create_report_table
{
	my ( $session ) = @_;
	
	return $session->get_db->do("CREATE TABLE `duplicates_report` (`eprintid_ref` int(11) NOT NULL default '0',`eprintid_dup` int(11) NOT NULL default '0',`score` int(11) NOT NULL default '0',`ignored` set('TRUE','FALSE') default NULL,PRIMARY KEY  (`eprintid_ref`, `eprintid_dup`)) ENGINE=MyISAM DEFAULT CHARSET=utf8 COMMENT='This table holds the suspected duplicates found after a global look-up';");
}


# save info in report table
sub save_duplicate
{
	my ( $session, $id_ref, $id_dup, $score ) = @_;

	# need to check first that the item doesn't already exist in the DB and set to IGNORED!!
	
	if( !does_exist( $session, $id_ref, $id_dup ) )
	{
		if($score <= 1)
		{
			$score = int($score *100);
		}

		my $sql = "INSERT INTO duplicates_report VALUES ( $id_ref, $id_dup, $score, \'FALSE\');";
		
		if( !$session->get_db->do( $sql ) )
		{
				# big problem!
		}
	}

	return;
}

# check if a pair (id_ref, id_dup) already exists and sets to IGNORED
sub does_exist
{
	my ( $session, $ref, $dup ) = @_;
	
	my $sql = "SELECT score FROM duplicates_report WHERE eprintid_ref='$ref' AND eprintid_dup='$dup';";
	
	my $sth = $session->get_db->prepare( $sql );
	
	if( $session->get_db->execute( $sth, $sql ) )
	{
		if( my @row = $sth->fetchrow )
		{
			return 1;
		}
		else
		{ 
			return 0;
		}
	}
	
	return 0;
}



# get them from the duplicates_report table (which was populated by another script...)
# NOT USED ANYMORE?
sub get_all_duplicates
{
	# if $include_ignored is set then we retrieve everything, otherwise the items not on the ignored list
	# if not set, it retrieves everything by default

	my ( $session, $include_ignored ) = @_;
	
	if(!defined $include_ignored)
	{
		$include_ignored = 1;
	}


# this parameter is now disabled!!
	$include_ignored = 1;


	my $sql;

	# global select:
	if($include_ignored)
	{
		$sql = "SELECT eprintid_ref, eprintid_dup, score, ignored FROM duplicates_report ORDER BY score DESC;";
	}
	else
	{
		# i know it's a bit silly to select "ignored" as we know this value will be == to FALSE... but it makes things more uniform in the rest of the script
		$sql = "SELECT eprintid_ref, eprintid_dup, score, ignored FROM duplicates_report WHERE ignored=\'FALSE\' ORDER BY score DESC;";
	}

	my $sth = $session->get_db->prepare( $sql );
	
	my @duplicates = [];
	
	if( $session->get_db->execute( $sth, $sql ) )
	{
		while( my @row = $sth->fetchrow )
		{
			
			push @duplicates, { eprintid_ref => $row[0], eprintid_dup => $row[1], score => $row[2], ignored => $row[3] };
			
		}
	}
	
	return \@duplicates;
		
}


# OK ;)
sub get_threshold
{
	return $EPrints::Duplicates::THRESHOLD;
}






# give it two fingerprints, it'll give you the score...
# the scoring method is encapsulated in this method so if one day the scoring method is modified, you won't have to modify the
# scripts calling it!
# NOT LONGER USED!
sub compare_fingerprints
{
	my ( $fp1, $fp2 ) = @_;
	
	return 0 if (! (defined $fp1 || defined $fp2) );
	
	return score($fp1, $fp2);
}


sub empty_fp_table
{
	my ( $session ) = @_;
	
	return $session->get_db->do( "TRUNCATE TABLE duplicates_archive_fingerprints" );
	
}

# obvious no?
sub create_fp_table
{
	my ( $session, $tablename ) = @_;
	
	return $session->get_db->do("CREATE TABLE `$tablename` (
			`eprintid` INT( 11 ) NOT NULL ,
			`fingerprint` TEXT NOT NULL ,
			PRIMARY KEY ( `eprintid` ) 
			) CHARSET=utf8 COMMENT = 'This table keeps cached fingerprints for the duplicate extensions.';"
		);	

}


# will generate all fingerprints for a given dataset
# WARNING: dataset feature currently disabled!
sub generate_all_fingerprints
{
	my ( $session ) = @_;

	my $ds = $session->get_archive()->get_dataset( "archive" );
	
	# retrieving ALL eprints from DATASET (could be done in one SQL statement...)
	my $searchexp = EPrints::SearchExpression->new(
		session => $session,
		dataset => $ds,
		allow_blank => 1,
	);
	
	$searchexp->perform_search();
	
	my $record_ids = $searchexp->get_ids();
	
	$searchexp->dispose;
	
	empty_fp_table( $session );
		
	my ($id, $fp, $sql);
	
	my $i = 0;
	
	foreach $id (@$record_ids)
	{
		my $eprint = EPrints::EPrint->new( $session, $id, $ds );
		
		next if(!defined $eprint);
	
		$fp = generate_fingerprint( $eprint );
		
		$sql = "INSERT INTO duplicates_archive_fingerprints VALUES ($id, ".$session->get_db->{dbh}->quote($fp).");";
		
		if( !$session->get_db->do( $sql ) )
		{
			# big problem!
		}
		else
		{
			$i++;
		}
		
	}

	return $i;	# return the number of generated fingerprints
}


# STILL UNTESTED
sub generate_incr_fingerprints
{
	# get all fingerprints we have
	# get all eprints from DB
	# generate FP for the eprints we don't have
	
	my ( $session ) = @_;

	my $ds = $session->get_archive()->get_dataset( "archive" );
	
	# retrieving ALL eprints from DATASET
	my $searchexp = EPrints::SearchExpression->new(
		session => $session,
		dataset => $ds,
		allow_blank => 1,
	);
	
	$searchexp->perform_search();
	
	my $record_ids = $searchexp->get_ids();
	
	$searchexp->dispose;
	
# now retrieve all the FPs we have	

	my $fps = get_all_fingerprints( $session );	# hash: $fps{eprintid} = fp
		
	my ($id, $fp, $sql);
	
	my $i = 0;
	
	foreach $id (@$record_ids)
	{
		next if defined( $$fps{$id} );
		
		my $eprint = EPrints::EPrint->new( $session, $id, $ds );
		
		next if(!defined $eprint);
	
		$fp = generate_fingerprint( $eprint );
		
		# next if($fp eq "")	?
	
		$sql = "INSERT INTO duplicates_archive_fingerprints VALUES ($id, ".$session->get_db->{dbh}->quote($fp).");";
		
		if( !$session->get_db->do( $sql ) )
		{
			# big problem!
		}
		else
		{
			$i++;
		}
		
	}

	return $i;	# return the number of generated fingerprints
}
	
	

# OK
sub get_fingerprint
{
	my ( $session, $eprintid ) = @_;

	my $fingerprint  = undef;

	return undef if ( !defined $eprintid );
	
	my $sql = "SELECT fingerprint FROM duplicates_archivet_fingerprints WHERE eprintid=$eprintid;";
	
	my $sth = $session->get_db->prepare( $sql );
	
	if( $session->get_db->execute( $sth, $sql ) )
	{
		my @row = $sth->fetchrow;

		if(defined $row[0])
		{
			$fingerprint = $row[0];
		}
	}

	return $fingerprint;

}



1;
